from enum import Enum


class CodeType(Enum):
    STOCK = 0
    SECTOR = 1

    def __str__(self):
        return self.name


class OutputType(Enum):
    SINGLE = 0
    MULTI = 1

    def __str__(self):
        return self.name

    def __index__(self):
        return self.value


class ExitType(Enum):
    # Return codes
    SUCCESS = 0
    FAILURE = -1
    # Hidden config
    IMPOSSIBLE = 712

    def __int__(self):
        return self.value


# Global variables
STOCK = CodeType.STOCK
SECTOR = CodeType.SECTOR

SINGLE = OutputType(0)
MULTI = OutputType(1)

SUCCESS = ExitType.SUCCESS
FAILURE = ExitType.FAILURE
IMPOSSIBLE = ExitType.IMPOSSIBLE



from collections import defaultdict

from pandas import to_datetime

from kiwoom.config import types
from kiwoom.config.const import MARKETS, MARKET_GUBUNS
from kiwoom.config.types import STOCK, SECTOR
from kiwoom.data.preps import number, string, remove_sign
from kiwoom.utils import list_wrapper


"""
Global Variables (Int, Dictionary type)
  - Each variable's contents may change from time to time
  - For type dict, update it periodically by following ways 
    1) config.global_variable[key] = val
    2) config.global_variable.update({key1: val1, ..., keyn: valn})
    3) del config.global_variable[key] 
"""
# Download configuration
SPEEDING = False
DISCIPLINED = False
REQUEST_LIMIT_TIME = 3600
REQUEST_LIMIT_TRY = float('inf')
REQUEST_LIMIT_ITEM = float('inf')


# Download progress bar divisor
DOWNLOAD_PROGRESS_DISPLAY = 10


# Code lengths for each type
SECTOR_CODE_LEN = 3
STOCK_CODE_LENS = [6, 8, 9]  # 일반주식 = 6, 하이일드펀드 = 8, 신주인수권 = 9


# Periods
PERIODS = ['tick', 'min', 'day', 'week', 'month', 'year']


# 체결시간 예외 케이스 변환기 (in Regular Expression)
EXCEPTIONAL_DATETIME_REPLACER = {
    '888888$': '160000',  # 장마감 시간에 이루어진 거래 (16:00:00)
    '999999$': '180000'  # 시간외 종료에 이루어진 거래 (18:00:00)
}


# 예외적인 코드의 예외적인 체결시간 케이스 변환기
EXCEPTIONAL_DATETIME_REPLACER_FOR_EXCEPTIONAL_CODE = {
    '253': {  # K200 F 매수 콜매도
        '888888$': '170000',
        '999999$': '180000',
    },
    '254': {  # K200 F 매도 풋매도
        '888888$': '170000',
        '999999$': '180000',
    }
}


# 장 시작시간이 변경된 예외
DELAYED_MARKET_OPENING = {
    # 'YYYYMMDD': delay(hour)
    '20201203': 1,  # 수능일 1시간 지연
    '20211118': 1,  # 수능일 1시간 지연
    '20221117': 1,  # 수능일 1시간 지연
    '20231116': 1,  # 수능일 1시간 지연
    '20241114': 1,  # 수능일 1시간 지연
    '20251113': 1,  # 수능일 1시간 지연
}


"""
Configuration for downloading historical data.
  - methods : 'is_~' and 'get_~' as helper methods
  - global variables : market, market_gubun, sector and etc.
  - private variables : used by 'is_~' and 'get_~' methods
"""


def is_date(date):
    date = str(date)
    if len(date) != len('YYYYMMDD'):
        print(f"'date' must be in form 'YYYYMMDD', not {date}.")
        return False
    try:
        datetime = to_datetime(date, format=_PERIOD_TO_DATETIME_FORMAT['day'])
    except Exception:  # TypeError, ValueError
        print(f"'date' must be in form 'YYYYMMDD' with appropriate value, not {date}.")
        return False
    return True


def is_market(code):
    code = str(code)
    if code not in MARKETS:
        # print(f'Market code must be one of {MARKETS}')
        return False
    return True


def is_market_gubun(code):
    code = str(code)
    if code not in MARKET_GUBUNS:
        # raise ValueError(f'Market Gubun code must be one of {MARKET_GUBUNS}')
        return False
    return True


def is_sector(code):
    code = str(code)
    if len(code) == 3:
        return True
    return False


def get_code_type(code):
    """
    Returns whether code belongs to stock or sector.
    """
    if len(code) == SECTOR_CODE_LEN:
        return SECTOR
    # elif len(code) in STOCK_CODE_LENS:
    #    return STOCK
    else:
        # raise ValueError(f'Given code {code} is not a stock code nor a sector code.')
        return STOCK


def get_tr_code(periods, ctypes=None):
    """
    Returns TR code for given code type and period.

    :param ctypes: 'stock', 'sector', or list of ctype
    :param periods: 'tick', 'min', 'day', 'week', 'month', 'year', or list of period
    """
    tr_codes = list()
    periods = list_wrapper(periods)
    ctypes = list_wrapper(ctypes)

    for period in periods:
        if ctypes is None:
            tr_codes.extend([_PERIOD_TO_TR_CODE[period][ctype] for ctype in types.CodeType])
            continue
        tr_codes.extend([_PERIOD_TO_TR_CODE[period][ctype] for ctype in ctypes])

    if len(tr_codes) == 1:
        return tr_codes[0]
    return tr_codes


def get_period(tr_code):
    """
    Returns period for given TR code.
    Period can be one of  'tick', 'min', 'day', 'week', 'month' or 'year'.
    """
    return _TR_CODE_TO_PERIOD[tr_code]


def get_record_name_for_its_name(tr_code):
    return _CODE_TYPE_TO_RECORD_NAME[_TR_CODE_TO_CODE_TYPE[tr_code]]


def get_datetime_column(period):
    return _PERIOD_TO_DATETIME_COLUMN[period]


def get_datetime_format(period):
    return _PERIOD_TO_DATETIME_FORMAT[period]


def boost():
    global SPEEDING, DISCIPLINED
    SPEEDING = True
    DISCIPLINED = False

    global REQUEST_LIMIT_TIME, REQUEST_LIMIT_TRY, REQUEST_LIMIT_ITEM
    REQUEST_LIMIT_TIME = 550
    REQUEST_LIMIT_TRY = 1000
    REQUEST_LIMIT_ITEM = 99


def regret():
    global SPEEDING, DISCIPLINED
    SPEEDING = False
    DISCIPLINED = True

    global REQUEST_LIMIT_TIME, REQUEST_LIMIT_TRY, REQUEST_LIMIT_ITEM
    REQUEST_LIMIT_TIME = 3600
    REQUEST_LIMIT_TRY = float('inf')
    REQUEST_LIMIT_ITEM = float('inf')


def preper(tr_code, otype):
    """
    Returns needed keys to fetch and pre-processor for each key as a tuple

    :param tr_code: str
        one of TR codes listed in KOA Studio or API Manual Guide
    :param otype: OutputType
        type can be either single or multi
    :return: tuple
        each element in tuple has key and pre-processor for its key, i.e. ((key1, function1), ...)
    """
    return ((key, _PREP_FOR_OUTPUTS[key]) for key in _OUTPUTS_FOR_TR_CODE[tr_code][otype])


def inputs(tr_code, code, unit=None, end=None):
    """
    Returns an iterator of key, val for each TR request

    :param tr_code: str
    :param code: str
    :param unit: int/str
    :param end: str
    :param prev_next: str
    :return: iterator
    """
    # Copy needed inputs to modify
    inputs = dict(_INPUTS_FOR_TR_CODE[tr_code])

    # To set code with appropriate record name for each TR code
    record_name = get_record_name_for_its_name(tr_code)  # '종목코드' or '업종코드'
    inputs[record_name] = code

    # To set key, val for each different period
    use_unit = get_tr_code(['tick', 'min'])
    use_date = get_tr_code(['day', 'week', 'month', 'year'])

    if tr_code in use_unit:
        inputs['틱범위'] = str(unit) if unit else '1'
    elif tr_code in use_date:
        inputs['기준일자'] = str(end) if end else ''  # cf. 기준일자 != start
    else:
        raise KeyError(f'Tr_code must be one of opt10079 ~ opt10083.')
    return inputs.items()


def outputs(tr_code, otype):
    """
    Returns needed keys to fetch data for each OutputType

    :param tr_code: str
        one of TR codes listed in KOA Studio or API Manual Guide
    :param otype: OutputType
        type can be either SINGLE or MULTI
    :return: list
    """
    return _OUTPUTS_FOR_TR_CODE[tr_code][otype]


"""
Protected Variables
  - Instead of directly accessing to these variables, use get methods in this module 
"""
# Map period, code type to matching TR code
_PERIOD_TO_TR_CODE = {
    'tick': {
        STOCK: 'opt10079',
        SECTOR: 'opt20004'
    },
    'min': {
        STOCK: 'opt10080',
        SECTOR: 'opt20005'
    },
    'day': {
        STOCK: 'opt10081',
        SECTOR: 'opt20006'
    },
    'week': {
        STOCK: 'opt10082',
        SECTOR: 'opt20007'
    },
    'month': {
        STOCK: 'opt10083',
        SECTOR: 'opt20008'
    },
    'year': {
        STOCK: 'opt10094',
        SECTOR: 'opt20019'
    }
}

# Map TR code to matching period
_TR_CODE_TO_PERIOD = {
    tr_code: period for period, tr_dic in _PERIOD_TO_TR_CODE.items() for ctype, tr_code in tr_dic.items()
}

# Map TR code to code type
_TR_CODE_TO_CODE_TYPE = {
    tr_code: ctype for tr_dic in _PERIOD_TO_TR_CODE.values() for ctype, tr_code in tr_dic.items()
}

# Map code type to record name to fetch code from downloaded data
_CODE_TYPE_TO_RECORD_NAME = {
    STOCK: '종목코드',
    SECTOR: '업종코드'
}

# Datetime column name for each period
_PERIOD_TO_DATETIME_COLUMN = {
    'tick': '체결시간',
    'min': '체결시간',
    'day': '일자',
    'week': '일자',
    'month': '일자',
    'year': '일자'
}

# Pandas parsing format of datetime column for each period
_PERIOD_TO_DATETIME_FORMAT = {
    # YYYYMMDDHHMMSS = %Y%m%d%H%M%S
    'tick': '%Y%m%d%H%M%S',
    'min': '%Y%m%d%H%M%S',
    'day': '%Y%m%d',
    'week': '%Y%m%d',
    'month': '%Y%m%d',
    'year': '%Y%m%d'
}


"""
Configuration for pre-process
"""
# How to pre-process for each output
__PREP_FOR_OUTPUTS = {
    number: [
        '평가손익', '총평가손익금액', '수익률(%)',
        '총수익률(%)', '주문가격', '주문번호',
        '주문수량', '예수금',  '보유수량',
        '매입가', '전일종가', '미체결수량',
        '체결량', '거래량', '거래대금'
    ],
    string: [
        '종목번호', '종목코드', '업종코드',
        '종목명', '주문상태', '일자',
        '체결시간', '주문구분'
    ],
    remove_sign: [
        '현재가', '시가', '고가', '저가'
    ]
}

# Revert dictionary to be in the form of {'key': function}
_PREP_FOR_OUTPUTS = defaultdict(
    lambda: string,
    {val: key for key, vals in __PREP_FOR_OUTPUTS.items() for val in vals}
)


"""
Configuration for inputs and outputs
"""
# Inputs needed for each TR code request
_INPUTS_FOR_TR_CODE = {
    'opt10079': {  # 주식틱차트조회요청
        '종목코드': None,
        '틱범위': None,
        '수정주가구분': '1'
    },
    'opt10080': {  # 주식분봉차트조회요청
        '종목코드': None,
        '틱범위': None,
        '수정주가구분': '1'
    },
    'opt10081': {  # 주식일봉차트조회요청
        '종목코드': None,
        '기준일자': None,
        '수정주가구분': '1'
    },
    'opt10082': {  # 주식주봉차트조회요청
        '종목코드': None,
        '기준일자': None,
        '끝일자': None,
        '수정주가구분': '1'
    },
    'opt10083': {  # 주식월봉차트조회요청
        '종목코드': None,
        '기준일자': None,
        '끝일자': None,
        '수정주가구분': '1'
    },
    'opt10094': {  # 주식년봉차트조회요청
        '종목코드': None,
        '기준일자': None,
        '끝일자': None,
        '수정주가구분': '1'
    },
    'opt20001': {  # 업종현재가요청
        '시장구분': None,
        '업종코드': None
    },
    'opt20002': {  # 업종별주가요청
        '시장구분': None,
        '업종코드': None
    },
    'opt20003': {  # 전업종지수요청
        '업종코드': None
    },
    'opt20004': {  # 업종틱차트조회요청
        '업종코드': None,
        '틱범위': None
    },
    'opt20005': {  # 업종분봉조회요청
        '업종코드': None,
        '틱범위': None
    },
    'opt20006': {  # 업종일봉조회요청
        '업종코드': None,
        '기준일자': None
    },
    'opt20007': {  # 업종주봉조회요청
        '업종코드': None,
        '기준일자': None
    },
    'opt20008': {  # 업종월봉조회요청
        '업종코드': None,
        '기준일자': None
    },
    'opt20009': {  # 업종현재가일별요청
        '시장구분': None,
        '업종코드': None
    },
    'opt20019': {  # 업종년봉조회요청
        '업종코드': None,
        '기준일자': None
    },
}

# Outputs for each TR code in the form of {'TR code': [[single data], [multi data]]}
_OUTPUTS_FOR_TR_CODE = {
    'opt10079': [  # 주식틱차트조회요청
        ['종목코드', '마지막틱갯수'],
        [
            '체결시간', '현재가', '거래량'
        ]
    ],
    'opt10080': [  # 주식분봉차트조회요청
        ['종목코드'],
        [
            '체결시간', '시가', '고가',
            '저가', '현재가', '거래량'
        ]
    ],
    'opt10081': [  # 주식일봉차트조회요청
        ['종목코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt10082': [  # 주식주봉차트조회요청
        ['종목코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt10083': [  # 주식월봉차트조회요청
        ['종목코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt10094': [  # 주식년봉차트조회요청
        ['종목코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt20004': [  # 업종틱차트조회요청
        ['업종코드'],
        [
            '체결시간', '현재가', '거래량'
        ]
    ],
    'opt20005': [  # 업종분봉조회요청
        ['업종코드'],
        [
            '체결시간', '시가', '고가',
            '저가', '현재가', '거래량'
        ]
    ],
    'opt20006': [  # 업종일봉조회요청
        ['업종코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt20007': [  # 업종주봉조회요청
        ['업종코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt20008': [  # 업종월봉조회요청
        ['업종코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
    'opt20019': [  # 업종년봉조회요청
        ['업종코드'],
        [
            '일자', '시가', '고가',
            '저가', '현재가', '거래량',
            '거래대금'
        ]
    ],
}



from os import getcwd
from os.path import join
from textwrap import dedent
from time import time
from traceback import format_exc

from pandas import read_csv, DateOffset
from PyQt5.QtTest import QTest
from PyQt5.QtWidgets import QApplication

from kiwoom import config
from kiwoom.config import history
from kiwoom.config.screen import Screen
from kiwoom.config.types import ExitType
from kiwoom.core.kiwoom import Kiwoom
from kiwoom.core.server import Server
from kiwoom.data.share import Share
from kiwoom.utils.general import *
from kiwoom.utils.manager import Downloader, timer


class Bot:
    def __init__(self, server=None):
        self.api: Kiwoom = Kiwoom()
        self.scr: Screen = Screen()
        self.share: Share = Share()

        # Connect server as a slot
        self.server: Server = server if issubclass(type(server), Server) else Server()
        self.server.init(bot=self, api=self.api, share=self.share)
        self.default_connect(server)

    def default_connect(self, server):
        self.api.set_connect_hook('on_receive_tr_data', 'rq_name')
        self.api.connect('on_receive_tr_data', signal=self.history, slot=self.server.history)
        self.api.connect('on_event_connect', signal=self.login, slot=self.server.login)
        self.api.connect('on_receive_msg', slot=self.server.on_receive_msg)

    @timer
    def login(self):
        self.api.comm_connect()
        self.api.loop()

    def connected(self):
        """
        Returns whether Bot is currently connected to server

        :return: bool
        """
        state = self.api.get_connect_state()
        if state == 1:
            return True
        return False

    def stock_list(self, market) -> list[str]:
        """
        Returns all stock codes in the given market

        :param market: str
            one of market code in kiwoom.config.markets
        """
        market = str(market)
        codes = self.api.get_code_list_by_market(market)
        codes = codes.split(';')[:-1]
        if not codes:
            raise ValueError(f'No stock codes for {market=}.')
        return sorted(codes)

    def sector_list(self, market_gubun) -> list[str]:
        """
        Returns all sector codes in the given market_gubun

        :param market_gubun: str
            one of market_gubun code in kiwoom.config.market_gubuns
        """
        market_gubun = str(market_gubun)
        if history.is_market_gubun(market_gubun):
            lst = list()
            sectors = self.api.koa_functions('GetUpjongCode', market_gubun)
            for sector in sectors.split('|')[:-1]:
                lst.append(sector.split(',')[1])
            return sorted(lst)

    def history(
            self,
            code,
            period,
            unit=None,
            start=None,
            end=None,
            path=None,
            merge=True,
            warning=True,
            prev_next='0'
    ):
        """
        Download historical market data of given code and save it as csv to given path

        :param code: str
            unique code of stock or sector
        :param period: str
            one of tick, min, day, week, month and year
        :param unit: int
            1, 3, 5, 10, 30 etc.. (cf. 1 bar = unit * period)
        :param start: str
            string of start day in format 'YYYYMMDD'
        :param end: str
            string of end day in format 'YYYYMMDD'. if None, until now by default.
        :param path: str
            path to save downloaded data
        :param merge: bool
            whether to merge data with existing file or to overwrite it
        :param warning: bool
            turn on/off the warning message if any
        :param prev_next: str
            this param is given by the response from the server. default is '0'
        """
        # Wait for default request limit, 3600 ms
        QTest.qWait(history.REQUEST_LIMIT_TIME)

        ctype = history.get_code_type(code)  # ctype = 'stock' | 'sector'
        tr_code = history.get_tr_code(period, ctype)

        """
            Setting args just for once.
        """
        if prev_next == '0':
            # In case path is '' or None
            if not path:
                path = getcwd()

            # To share variables with Slot
            kwargs = effective_args(locals(), remove=['ctype', 'tr_code'])
            self.share.remove_single(name())
            self.share.update_single(name(), 'error', False)
            self.share.update_single(name(), 'restart', False)
            self.share.update_single(name(), 'complete', False)
            self.share.update_single(name(), 'impossible', False)

            # To check format of input dates
            if 'start' in kwargs:
                if not history.is_date(start):
                    raise ValueError(f"Given 'start' {start} is not a valid date.")
            if 'end' in kwargs:
                if not history.is_date(end):
                    raise ValueError(f"Given 'end' {end} is not a valid date.")

            """
                Check 'start' and 'end' points to save downloading time. 
            """
            if merge:
                try:
                    file = join(path, code + '.csv')
                    col = history.get_datetime_column(period)
                    df = read_csv(
                        file,
                        index_col=[col],
                        parse_dates=[col],
                        encoding=config.ENCODING
                    )

                    if period in ['tick', 'min']:
                        # Last tick for stock is 15:30 and for sector is 18:00
                        h, m = (15, 30) if ctype is history.STOCK else (18, 00)  # else for sector
                        last_day = date(df.index[-1])
                        last_tick_of_day = Timestamp(df.index[-1]).replace(hour=h, minute=m)
                        download_completed = last_tick_of_day <= df.index[-1]

                        # To push 'start' date further as much as possible. If None, set newly.
                        if 'start' not in kwargs or date(kwargs['start']) <= last_day:
                            if download_completed:
                                # Start from the day after last day
                                kwargs['start'] = str((last_day + DateOffset(1)).date()).replace('-', '')
                            else:
                                # Start from the last day
                                kwargs['start'] = str(last_day).replace('-', '')

                        # If downloading is not needed, just return
                        if 'end' in kwargs:
                            if download_completed:
                                if date(kwargs['end']) <= last_day:
                                    self.share.update_single(name(), 'complete', True)
                                    return

                    else:  # if period in ['day', 'week', 'month', 'year']
                        last_day = date(df.index[-1])
                        # To push 'start' date further as much as possible. If None, set newly.
                        if 'start' not in kwargs or date(kwargs['start']) <= last_day:
                            # Start from the last day
                            kwargs['start'] = str(last_day).replace('-', '')

                        # If downloading is not needed, just return
                        if 'end' in kwargs:
                            if date(kwargs['end']) < last_day:
                                self.share.update_single(name(), 'complete', True)
                                return

                    # Once read, use later in Server.history_to_csv() to increase efficiency
                    self.share.update_single(name(), 'file', df)

                # If any exception, just skip
                except Exception as err:
                    pass

            """
                Update and print arguments. 
            """
            # Done arg setting
            self.share.update_args(name(), kwargs)

            # Print args
            f = lambda key: f"'{kwargs[key]}'" if key in kwargs else None
            print(f"{{code={f('code')}, start={f('start')}, end={f('end')}, period={f('period')}}}")

        """
            Start downloading.
        """
        # Check requesting status
        if 'histories' in self.share.single:
            self.share.single['histories']['nrq'] += 1
            if history.SPEEDING:
                if self.share.get_single('histories', 'nrq') >= history.REQUEST_LIMIT_TRY:
                    # Set back to default configuration
                    if self.share.get_single('histories', 'cnt') == 0:
                        self.share.update_single(name(), 'impossible', True)
                    self.share.update_single(name(), 'restart', True)
                    self.api.unloop()
                    return

        # Finally request data to server
        for key, val in history.inputs(tr_code, code, unit, end):
            self.api.set_input_value(key, val)
        scr_no = self.scr.alloc(tr_code, code)

        # If comm_rq_data returns non-zero error code, restart downloading
        if self.api.comm_rq_data(name(), tr_code, prev_next, scr_no) != 0:
            self.share.update_single(name(), 'impossible', True)
            self.share.update_single(name(), 'restart', True)
            self.api.unloop()
            return

        # Wait response from the server
        self.api.loop()

    @Downloader.watcher
    def histories(
            self,
            market=None,
            sector=None,
            period=None,
            unit=None,
            start=None,
            end=None,
            slice=None,
            code=None,
            path=None,
            merge=False,
            warning=True
    ):
        """
        Download historical data of partial or all items in given market/sector and save it as csv file.

        :param market: str
            one of market type in string
        :param sector: str
            one of market gubuns in string
        :param period: str
            one of tick, min, day, week, month and year
        :param unit: int
            1, 3, 5, 10, 30 etc.. (cf. 1 bar = unit * period)
        :param start: str
            string of start day in format 'YYYYMMDD'
        :param end: str
            string of end day in format 'YYYYMMDD'. if None, until now by default.
        :param path: str
            path to save downloaded data
        :param slice: tuple of int
            partially download from the whole items in specific market.
            slice can be one of (from, to), (from, None) or (None, to)
        :param code: str
            unique code of stock or sector to start downloading from.
        :param merge: bool
            whether to merge data with existing file or to overwrite it
        :param warning: bool
            turn on/off the warning message if any

        :return: int or tuple
            if successfully download all, returns 0 (= ExitCode.success)
            if download failed by local errors, returns -1 (= ExitCode.failure)
            if download restart is needed, returns slice that can be used in the next run
        """
        if not path:
            path = getcwd()

        # Initialize status
        self.share.remove_single(name())
        self.share.single[name()]['nrq'] = 0  # number of request
        self.share.single[name()]['cnt'] = 0  # number of stocks
        self.share.update_args(name(), effective_args(locals()))

        """
            Validate given arguments
        """
        # To decide what to download
        lst, ctype, mname = None, None, None
        if not any([market, sector]) or all([market, sector]):
            raise RuntimeError("Download target must be either of 'market' or 'sector'.")
        elif market is not None:
            market = market.upper()
            ctype, mname = str(history.STOCK).lower(), history.MARKETS[market]
            if market != 'NXT':
                krx = set(self.stock_list(market))
                nxt = set(self.stock_list('NXT'))
                nxt.intersection_update(krx)
                sor = set([c + '_AL' for c in list(nxt)])
                lst = (krx - nxt) | sor
                lst = sorted(list(lst))
            else:
                lst = self.stock_list(market)
                lst = [c + '_AL' for c in lst]
        elif sector is not None:
            ctype, mname = str(history.SECTOR).lower(), history.MARKET_GUBUNS[sector]
            lst = self.sector_list(sector)
        
        # Set the portion in download list
        from_, to_ = 0, None
        if all([slice, code]):
            raise RuntimeError("Only one option is available: either of 'slice' or 'start_code'.")
        # Option1 - Slice
        elif slice is not None:
            try:
                from_, to_ = slice
            except ValueError:
                raise ValueError(f'Slice must be (from, to), (from, None), or (None, to) not {slice}.')
        # Option2 - Code
        elif code is not None:
            from_, to_ = lst.index(code), None

        # Select target in download list
        tot = len(lst)
        lst = lst[from_: to_]

        # To print progress bar
        divisor = history.DOWNLOAD_PROGRESS_DISPLAY
        if len(lst) < 4 * divisor:
            # At least print 25% of progress if possible
            divisor = max(1, len(lst) // 4)

        """
            Start downloading
        """
        # To Download
        status = ''
        begin = time()
        print(f'Download Start for {len(lst)} {ctype}s in {mname}.')
        print(f' - Encoding : {config.ENCODING}\n - DataPath : {path}')

        for i, code in enumerate(lst):
            if i % divisor == 0:
                pct = ((from_ + i) / tot) * 100
                print(f'Downloading ..\t{pct: .1f}% ({from_ + i} of {tot})')

            # Try downloading
            try:
                self.history(code, period, unit=unit, start=start, end=end, path=path, merge=merge, warning=warning)

            # 1) Error with starting Bot.history() (at the first call)
            except Exception:
                args = unpack_args(self.share.get_args('history'))
                print(f"\nAn error at Bot.history({args}).\n\n{format_exc()}")
                return ExitType.FAILURE

            # 2) Error with continuing Signal.history() or Slot.history()
            if self.share.get_single('history', 'error'):
                # Note that error message will be printed
                return ExitType.FAILURE

            # 3) Error with reaching the request limit or error with frozen server
            elif self.share.get_single('history', 'restart'):
                # If it's impossible to download with the trick
                if self.share.get_single('history', 'impossible'):
                    print(f"\n[{clock()}] The {ctype} {code} can't be downloaded with speeding.")
                    return ExitType.IMPOSSIBLE
                break

            # 4) Unexpected error from Signal.history() or Slot.history()
            elif not self.share.get_single('history', 'complete'):
                print(f'\n[{clock()}] Stop downloading due to unexpected error at code {code}.\n')
                return ExitType.FAILURE

            """
                Download completed for one item in the list
            """
            # Finally successfully downloaded
            self.share.single[name()]['cnt'] += 1

            # 5) Successfully downloaded with disciplined, but it's time for speeding again.
            if history.DISCIPLINED:
                status = f"[{clock()}] The program needs to be restarted for speeding again."
                break

            # 6) Successfully downloaded but exceeds request limit items
            if history.SPEEDING:
                if self.share.single[name()]['cnt'] >= history.REQUEST_LIMIT_ITEM:
                    # To check whether items are downloaded by the actual requests to download more if possible
                    if self.share.single[name()]['nrq'] >= history.REQUEST_LIMIT_ITEM:
                        break

        """
            Close downloading
        """
        cnt = self.share.get_single(name(), 'cnt')
        cum = from_ + cnt
        msg = dedent(
            f"""
            Download Done for {100 * cum / tot if lst else 100: .1f}% ({cum} of {tot}) {ctype}s in {mname}.
            Download Time : {(time() - begin) / 60: .1f} minutes (with {self.share.single[name()]['nrq']} requests)\n
            """
        ) + status
        print(msg)

        # If complete
        if cnt == len(lst):
            return ExitType.SUCCESS
        # Else return remaining items
        return cum, to_

    def exit(self, ecode=0):
        """
        Close all windows open and exit the application that runs this bot

        :param ecode: int
            return code when exiting the program
        """
        app = QApplication.instance()
        app.closeAllWindows()
        app.exit(ecode)



from datetime import datetime
from sys import _getframe

from numpy import ndarray
from pandas import Series, Timestamp


# Global Variables
ARRAY_TYPE = (list, tuple, ndarray, Series)


def name():
    """
    Please use inside of a signal/slot method

    :return: the name of the caller, i.e. signal/slot function name
    """
    return _getframe(1).f_code.co_name


def clock():
    """
    Returns current time in format 'HH:MM' of string

    :return: str
    """
    now = datetime.now()
    return f'{now.hour}:{now.minute:02}'


def date(raw):
    """
    :param str:
    :return: datetime.date
    """
    return Timestamp(raw).date()


def list_wrapper(item):
    if item is None:
        return None
    if isinstance(item, ARRAY_TYPE):
        return item
    return [item]


def effective_args(local, remove=None):
    dic = dict(local)

    # Delete 'self' and non effective args
    for param, arg in local.items():
        if arg is None or arg == '':
            del dic[param]
        if param == 'self':
            del dic[param]

    # Delete unwanted args in local
    if remove is not None:
        if isinstance(remove, ARRAY_TYPE):
            for param in remove:
                del dic[param]
        else:
            del dic[remove]

    return dic


def unpack_args(local):
    if 'self' in local:
        del local['self']

    s = ''
    for arg, param in local.items():
        s += str(arg) + '=' + str(param) + ', '
    s = s[:-2] if s else s  # remove last comma and space ', '
    return s



"""
Default configuration of how to pre-process raw inputs from the server.
Default way to deal with data is considering one as a string.  

1) prep (default)
    : returns number(x) if possible, else string(x)

2) number
    : returns an int or a float

3) string
    : returns a string with white space removed

4) remove_sign (special case)
    : returns a number/string without any signs (+/-)
"""


def prep(x):
    """
    Preprocess x into number or string (default preprocessor)

    :param x: str
    :return: int/float/str
    """
    try:
        return number(x)
    except ValueError:
        return string(x)


def number(x):
    """
    First, tries to type-cast x into float if it has '.' dot in x. 
    If it fails, move on to int and if fails again retry float.
    If converting to float raises an ValueError, then throws the error.
    """
    if x is None or x.strip() == '':
        return None

    if '.' in x:
        try:
            return float(x)
        except ValueError:
            raise ValueError(f"{x} of type {type(x)} can't be a number.")

    try:
        return int(x)
    except ValueError:
        try:
            return float(x)
        except ValueError:
            raise ValueError(f"{x} of type {type(x)} can't be a number.")


def string(x):
    """
    Returns a string with white space removed.
    """
    return x.strip()


def remove_sign(x):
    """
    Returns number or string with '+' and '-' signs removed.

    1) Removes '+' or '-' from x
    2) Tries to type-cast x into number(int / float)
    3) If step number 2 fails, returns string.
    """
    x = x.translate({ord('+'): '', ord('-'): ''})
    try:
        x = number(x)
    except ValueError:
        x = string(x)
    return x



from functools import wraps


# Dictionary for error information
ERR = {
    0: ('OP_ERR_NONE', '정상처리'),
    -10: ('OP_ERR_FAIL', '실패'),
    -100: ('OP_ERR_LOGIN', '사용자정보교환실패'),
    -101: ('OP_ERR_CONNECT', '서버접속실패'),
    -102: ('OP_ERR_VERSION', '버전처리실패'),
    -103: ('OP_ERR_FIREWALL', '개인방화벽실패'),
    -104: ('OP_ERR_MEMORY', '메모리보호실패'),
    -105: ('OP_ERR_INPUT', '함수입력값오류'),
    -106: ('OP_ERR_SOCKET_CLOSED', '통신연결종료'),
    -200: ('OP_ERR_SISE_OVERFLOW', '시세조회과부하'),
    -201: ('OP_ERR_RQ_STRUCT_FAIL', '전문작성초기화실패'),
    -202: ('OP_ERR_RQ_STRING_FAIL', '전문작성입력값오류'),
    -203: ('OP_ERR_NO_DATA', '데이터없음'),
    -204: ('OP_ERR_OVER_MAX_DATA', '조회가능한종목수초과'),
    -205: ('OP_ERR_DATA_RCV_FAIL', '데이터수신실패'),
    -206: ('OP_ERR_OVER_MAX_FID', '조회가능한FID수초과'),
    -207: ('OP_ERR_REAL_CANCEL', '실시간해제오류'),
    -211: ('NO_OFFICIAL_ERROR_NAME', '계좌조회과부하'),
    -300: ('OP_ERR_ORD_WRONG_INPUT', '입력값오류'),
    -301: ('OP_ERR_ORD_WRONG_ACCTNO', '계좌비밀번호없음'),
    -302: ('OP_ERR_OTHER_ACC_USE', '타인계좌사용오류'),
    -303: ('OP_ERR_MIS_2BILL_EXC', '주문가격이20억원을초과'),
    -304: ('OP_ERR_MIS_5BILL_EXC', '주문가격이50억원을초과'),
    -305: ('OP_ERR_MIS_1PER_EXC', '주문수량이총발행주수의1%초과오류'),
    -306: ('OP_ERR_MIS_3PER_EXC', '주문수량은총발행주수의3%초과오류'),
    -307: ('OP_ERR_SEND_FAIL', '주문전송실패'),
    -308: ('OP_ERR_ORD_OVERFLOW', '주문전송과부하'),
    -309: ('OP_ERR_MIS_300CNT_EXC', '주문수량300계약초과'),
    -310: ('OP_ERR_MIS_500CNT_EXC', '주문수량500계약초과'),
    -340: ('OP_ERR_ORD_WRONG_ACCTINFO', '계좌정보없음'),
    -500: ('OP_ERR_ORD_SYMCODE_EMPTY', '종목코드없음')
}


def msg(ecode):
    """
    Returns an error message that corresponds to given error code.

    Defined errors will be found in config.error.err (dict)

    :param ecode: int
        One of error code defined at err dictionary.
    :return: str
        One line error message combined with code, type, and explanation.
    """
    etype, msg = ERR[ecode]
    if ecode == 0:
        return f'{msg}'
    return f'Error {{Code: {ecode}, Type: {etype}, Msg: {msg}}}'


def catch_error(fn):
    """
    Decorator function that catches an error from the returns of some api methods.

    If an api method that requests to the server returns the result of request as an error code,
    this decorator function checks the error code. If error code is other than 0, this function
    prints the error message corresponding to the error code. It does nothing, otherwise.

    :param fn: method
        One of API methods that returns the error code as result.
    :return: function
        Function that wraps given fn and checks the output of fn.
    """
    @wraps(fn)  # keep docstring of original function, fn.
    def wrapper(*args, **kwargs):
        out = fn(*args, **kwargs)
        if out != 0:  # 0: ('OP_ERR_NONE', '정상처리')
            print(f'\nAn error occurred from {fn.__name__}.\n  * {msg(out)}')
        return out
    return wrapper



# Version
from .__version__ import __version__

# Namespace Control
from . import (
    config,
    core,
    data,
    utils,
    wrapper
)

from .core import Bot, Server, Kiwoom
from .wrapper import API
